package scala.events

import scala.language.implicitConversions
import scala.collection.mutable.{ListBuffer,Stack}
import scala.util.DynamicVariable

import scala.actors.Actor
import scala.actors.Actor._


trait Event[+T] {
  /*
   * Sink function type: takes event id, event parameter
   * fills the given list with reactions to be executed
   * the reactions represt
   */
  type Sink = (Int, T, Boolean, ListBuffer[(() => Unit, Trace)]) => Unit
  type Trace = List[Event[_]]

  /*
   * Register a sink function to the event (used for propagation of events)
   */
  protected[events] def +=(sink: Sink)

  /*
   * Unregister a sink function
   */
  protected[events] def -=(sink: Sink)

  /*
   * Register a reaction to the event
   */
  def +=(react: T => Unit)

  /*
   * Unregister a reaction
   */
  def -=(react: T => Unit)

  /**
   * Events disjunction.
   */
  def ||[S >: T, U <: S](other: Event[U]) = new EventNodeOr[S](this, other)


  /**
   * Creates an events sequence
   */
  def then[U, V, S >: T](other: => Event[U], merge: (S, U) => V) = new EventNodeSequence[S, U, V](this, other, merge)

  /**
   * Creates an events sequence with a merge method creating a tuple of both event parameters
   */
  def then[U, S >: T](other: => Event[U]) = new EventNodeSequence[S, U, (S, U)](this, other, (p1: S, p2: U) => (p1, p2))

  /**
   * Event filtered with a predicate
   */
  def &&[U >: T, V >: T <: U](pred: U => Boolean) = new EventNodeFilter[V](this, pred)

  /**
   * Event filtered with a boolean variable
   */
  def &&[U >: T](pred: () => Boolean) = new EventNodeFilter[U](this, _ => pred())

  
  /**
   * Event is triggered except if the other one is triggered
   * @TODO what is the semantics of this in concurrent programs?
   */
  def \[U >: T,S](other: Event[S]) = new EventNodeExcept[U,S](this, other)

  /**
   * Events conjunction
   */
  def and[U, V, S >: T](other: Event[U], merge: (S, U) => V) = new EventNodeAnd[S, U, V](this, other, merge)

  /*
  * Event conjunction with a merge method creating a tuple of both event parameters
  */
  def and[U, S >: T](other: Event[U]) = new EventNodeAnd[S, U, (S, U)](this, other, (p1: S, p2: U) => (p1, p2))

  /**
   * Transform the event parameter
   */
  def map[U, S >: T, V >: T <: S](mapping: S => U) = new EventNodeMap[V, U](this, mapping)

  /**
   * Transform the event parameter independently of the parameter value
   */
  def map[U, S >: T](mapping: () => U) = new EventNodeMap[S, U](this, _ => mapping())

  /**
   * Drop the event parameter; equivalent to map((_: Any) => ())
   */
  def dropParam[S >: T] = new EventNodeMap[S, Unit](this, _ => ())

  def !![T](ev:Event[T]) = new EventNodeDoubleBang(ev)


}

/*
 *  Base class for events.
 *  Maintains sinks and reactions and
 */
abstract class EventNode[T] extends Event[T] {
  protected val sinks = new ListBuffer[Sink]
  protected[events] val _reactions = new ListBuffer[T => Unit]
  protected val definer = {
    val t = Thread.currentThread.getStackTrace.toList
    val sf = t.lastIndexWhere(x => x.getClassName.startsWith("scala.events")) + 1
    t(sf).getFileName + ":" + t(sf).getLineNumber
  }
  /*
   * Register a reaction to the event
   */
  def +=(sink: Sink) {
    sinks += sink
    // deploy the event if the first reaction/sink is registered
    if (sinks.size == 1 && _reactions.size == 0)
      deploy
  }


  def -=(sink: Sink) {
    sinks -= sink
    // undeploy the event if the last reaction/sink is unregistered
    // has cascade effect
    // avoids redundant event propagation
    if (sinks.size == 0 && _reactions.size == 0)
      undeploy
  }

  /*
   * Register a reaction to the event
   */
  def +=(react: T => Unit) = _reactions.synchronized {
    _reactions += react
    // deploy the event if the first reaction/sink is registered
    if (_reactions.size == 1 && sinks.size == 0)
      deploy
  }

  /*
   * Unregister a reaction
   */
  def -=(react: T => Unit) =  _reactions.synchronized {
    _reactions -= react
    // undeploy the event if the last reaction/sink is unregistered
    if (sinks.size == 0 && _reactions.size == 0)
      undeploy
  }

  /**
   * Deploy the event (i.e. registers to the referenced events)
   */
  protected def deploy: Unit

  /**
   * Undeploy the event (i.e. unregisters from the referenced events)
   */
  protected def undeploy: Unit

  /** Collects the reactions registered with this event and associates the current event trace.
   *  It then propagates to sinks.
   */
  protected[events] def reactions(id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) {
    eventTrace.withValue(this :: eventTrace.value) {
      _reactions.synchronized {
	// collect the reactions registered with this event
	_reactions.foreach(react => reacts += ((() => react(v)) -> eventTrace.value))
      }
      // propagate to sinks adding this event to the event trace
      sinks.foreach(sink => sink(id, v, sync, reacts))
    }
  }

}

protected[events] object eventTrace extends DynamicVariable[List[Event[_]]](Nil)

object EventIds {
  // last used event id
  var lastId: Int = 0

  /*
  * generate new event id
  */
  def newId(): Int = synchronized {
    lastId += 1
    lastId
  }
}

/*
 * An implementation of an explicit event
 */
abstract class AbstractImperativeEvent[T] extends EventNode[T] {
  private var deployed = false
  /*
  * Trigger the event
  */
  def triggerEvent(v: T,sync:Boolean) = {
    beforeTrigger(v)
    // does something only if the event is deployed, i.e. if some reactions or sinks
    // are registered
    if(deployed) {
      // collect matching reactions
      val reacts: ListBuffer[(() => Unit, Trace)] = new ListBuffer
      reactions(EventIds.newId(), v, sync , reacts)
      // once reactions are collected, we are after the triggering
      afterTrigger(v)
      // execute the collected reactions
      if (!reacts.isEmpty)
	executeReactions(reacts,sync)
    } else {
      afterTrigger(v)
    }
  }


  def executeReactions(reacts:ListBuffer[(() => Unit, Trace)],sync:Boolean)=reacts.foreach(
         (r:(()=>Unit,Trace)) => {
           val react=r._1
           val trace=r._2
           eventTrace.withValue(trace) {
               react()
           }
         }
       )
  


  protected[events] def beforeTrigger(v: T) {}
  protected[events] def afterTrigger(v: T) {}

  protected override def deploy { deployed = true }

  protected override def undeploy { deployed = false }

  override def toString = getClass.getName
  def apply(v:T)
}

class ImperativeEvent[T] extends AbstractImperativeEvent[T] {
  def apply(v:T)=triggerEvent(v,true)
}



object AsyncProvider {
  private val countProviders=2
  private var providerType=1;
  def setProviderType(id:Int)={providerType=(id%countProviders)}
  def getProviderType=providerType
  def create:AsyncProvider={
    getProviderType match {
      case 0=>new AsyncProvider_FullThread()
      case 1=>new ASyncProvider_Actor()
    }
  }
}

abstract class AsyncProvider {
  def apply(task : ()=>Unit)
  def haltAsync()
}

// serialize occurrences
class ASyncProvider_Actor extends AsyncProvider {
  private case class handleAsync(task : ()=>Unit)
  private case class Halt() {}
  lazy val myActor=new Actor {
    var active=false
    def act={
      active=true
      var halting:Boolean=false
      loopWhile (!halting) {
	react {
	  case handleAsync(task)=> task()
	  case Halt=>{
	    halting=true
	  }
	}
      }
      active=false
    }
  }
  def apply(task : ()=>Unit)={
    if (!myActor.active)
      myActor.start
    myActor ! handleAsync(task)
  }
  def haltAsync()={
    if (myActor.active) {
      myActor ! Halt
    }
  }
}

// allow handling multiple occurences
class AsyncProvider_FullThread extends AsyncProvider {
  class Worker(task :()=>Unit) extends Thread {
    override def run()={
      task()
    }
  }
  def apply(task : ()=>Unit)={
    val w=new Worker(task)
    w.start
  }
  def haltAsync()={
  }
}



// chose the type of Async behavior for all async events (1 or m)
class ImperativeAsyncEvent[T] extends AbstractImperativeEvent[T]  {
  val sync:Boolean=false
  val asyncProvider=AsyncProvider.create
  def apply(v:T)=triggerEvent(v,false)

  override def executeReactions(reacts:ListBuffer[(() => Unit, Trace)],sync:Boolean)={
    if (sync) {
      super.executeReactions(reacts,sync)
    } else {
      asyncProvider(()=>super.executeReactions(reacts,sync))
    }
  }

  def haltAsync()={
    asyncProvider.haltAsync()
  }

  override def undeploy {
    haltAsync()
    super.undeploy
  }

}

class EventNodeDoubleBang[T](ev:Event[T]) extends EventNode[T] {
  var asyncProvider:Option[AsyncProvider]=None
  lazy val onEvt = (id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    if (!sync) {
      eventTrace.withValue(this :: eventTrace.value) {
	reactions(id, v, false, reacts)
	}
    } else {
      val async_reacts: ListBuffer[(() => Unit, Trace)] = new ListBuffer
      reactions(EventIds.newId(), v, false , async_reacts)
      if (!async_reacts.isEmpty) {
	eventTrace.withValue(this :: eventTrace.value) {
	  reacts += ((()=>delayed_handler(async_reacts)) -> eventTrace.value)
	}
      }
    }
  }:Unit

  // to execute collected handlers from this node on asynchronous, this delayed
  // handler is executed together with normal handlers but will execute the collected handlers 
  def delayed_handler(collected_reactions:ListBuffer[(() => Unit, Trace)])():Unit={
    asyncProvider match {
      case Some(_)=>{}
      case None=>{
	asyncProvider=Some(AsyncProvider.create)
      }
    }

    asyncProvider match {
      case Some(asyncprov)=>{
	asyncprov(()=>{collected_reactions.foreach(
        (r:(()=>Unit,Trace)) => {
          val react=r._1
          val trace=r._2
          eventTrace.withValue(trace) {
            react()
          } } )
		       })
      }
      case None=>{}
    }
  }



  protected override def deploy {
    ev += onEvt
  }

  /*
  * Unregister from the referenced events
  */
  protected override def undeploy {
    ev -= onEvt
    asyncProvider match {
      case Some(ap) => {
	ap.haltAsync()
	asyncProvider=None
      }
      case None =>{}
    }
  }

  override def toString = "!!(" + ev + ")"
}


/*
 * Implementation of event conjunction
 */
class EventNodeAnd[T1, T2, T](ev1: Event[T1], ev2: Event[T2], merge: (T1, T2) => T) extends EventNode[T] {

  // The id of the last received event
  var id: Int = 0

  // Parameter value of event1 or event2 (depending on which is received first)
  var v1: T1 = _
  var v2: T2 = _
  var currentTrace: Trace = Nil

  /*
  * Reaction to event1
  */
  lazy val onEvt1 = (id: Int, v1: T1, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    if (this.id == id) {
      // event2 is already received; collect the reactions
      eventTrace.withValue(currentTrace ::: eventTrace.value) {
        reactions(id, merge(v1, this.v2), sync, reacts)
      }
    }
    else {
      // event2 is not received yet; save the data of the event1
      this.id = id
      this.v1 = v1
      this.currentTrace = eventTrace.value
    }
  }

  /*
   * Reaction to event2
   */
  lazy val onEvt2 = (id: Int, v2: T2, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    if (this.id == id) {
      // event1 is already received; collect the reactions
      eventTrace.withValue(currentTrace ::: eventTrace.value) {
        reactions(id, merge(this.v1, v2), sync,  reacts)
      }
    }
    else {
      // event1 is not received yet; save the data of the event2
      this.id = id
      this.v2 = v2
      this.currentTrace = eventTrace.value
    }
  }

  /*
  * Register to the referenced events
  */
  protected override def deploy {
    ev1 += onEvt1
    ev2 += onEvt2
  }

  /*
  * Unregister from the referenced events
  */
  protected override def undeploy {
    ev1 -= onEvt1
    ev2 -= onEvt2

  }

  override def toString = "(" + ev1 + " and " + ev2 + ")"

}


class EventMonitor[T](ev1:Event[T]) extends EventNode[T] {
  lazy val onEvt = (id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    println("Monitor Event arg:"+v+" sync:"+sync+ " @"+definer)
    reactions(id, v, sync, reacts)
  }
  protected override def deploy {
  }
  protected override def undeploy {
  }
  ev1 += onEvt
  override def toString = "Monitor(" + ev1 + ")"
}

/*
 * Implementation of event disjunction
 */
class EventNodeOr[T](ev1: Event[_ <: T], ev2: Event[_ <: T]) extends EventNode[T] {

  // The id of the last received event
  var id: Int = 0

  /*
   * Reaction to both events
   */
  lazy val onEvt = (id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    // if the event was already received, avoid processing it twice
    if (this.id != id) {
      this.id = id
      // the event received for the first time; collect the reactions
      reactions(id, v, sync, reacts)
    }
  }

  /*
  * Register to the referenced events
  */
  protected override def deploy {
    ev1 += onEvt
    ev2 += onEvt
  }

  /*
  * Unregister from the referenced events
  */
  protected override def undeploy {
    ev1 -= onEvt
    ev2 -= onEvt
  }

  override def toString = "(" + ev1 + " || " + ev2 + ")"
}

/*
 * Implements transformation of event parameter
 */
class EventNodeMap[T, U](ev: Event[T], f: T => U) extends EventNode[U] {

  /*
   * Reaction to the referenced event
   */
  lazy val onEvt = (id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    // transform v to f(v)
    reactions(id, f(v), sync, reacts)
  }

  /*
  * Register to the referenced events
  */
  protected override def deploy {
    ev += onEvt
  }

  /*
  * Unregister from the referenced events
  */
  protected override def undeploy {
    ev -= onEvt
  }

  override def toString = getClass.getName

}

/*
 * Implements filtering event by a predicate
 */
class EventNodeFilter[T](ev: Event[T], f: T => Boolean) extends EventNode[T] {

  /*
   * Reaction to the referenced event
   */
  lazy val onEvt = (id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    // filter the event by f(v)
    if (f(v)) {
      reactions(id, v, sync, reacts)
    }
  }

  /*
  * Register to the referenced events
  */
  protected override def deploy {
    ev += onEvt
  }

  /*
  * Unregister from the referenced events
  */
  protected override def undeploy {
    ev -= onEvt
  }

  override def toString = "(" + ev + " && <predicate>)"

}


/*
 * Implementation of quantification over a (varying) list of objects
 */
class EventNodeExists[T, U](list: VarList[T], evf: T => Event[U]) extends EventNode[U] {

  /*
   * Register to the event of a newly added list element
   */
  lazy val onElementAdded = (target: T) => {
    evf(target) += onEvt
  }

  /*
   * Unregister from the event of a removed list element
   */
  lazy val onElementRemoved = (target: T) => {
    evf(target) -= onEvt
  }

  /*
   * Reaction to the observed events
   */
  lazy val onEvt = (id: Int, v: U, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    reactions(id, v, sync, reacts)
  }

  /*
  * Register to the events of all list elements and the list changes
  */
  protected override def deploy {
    list.foreach(target => evf(target) += onEvt)
    list.elementAdded += onElementAdded
    list.elementRemoved += onElementRemoved
  }

  /*
  * Unregister from the events of all list elements and the list changes
  */
  protected override def undeploy {
    list.foreach(target => evf(target) -= onEvt)
    list.elementAdded -= onElementAdded
    list.elementRemoved -= onElementRemoved
  }

  override def toString = getClass.getName

}



/*
 * Implementation of event sequence operator
 */
class EventNodeSequence[T, U, V](ev1: Event[T], ev2: => Event[U], merge: (T, U) => V) extends EventNode[V] {

  // the id of the last received event1
  // -1 if event1 was not received yet (or after last event2)
  var id: Int = -1
  // value of the last event1
  var v1: T = _

  /*
  * Reaction to event1
  */
  lazy val onEvt1 = (id: Int, v1: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    // ignore consecutive occurrences of event1
    if (this.id == -1) {
      // save the data of event1
      this.id = id
      this.v1 = v1
    }
  }

  /*
   * Reaction to event2
   */
  lazy val onEvt2 = (id: Int, v2: U, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    // react to event2 only if event1 was already received;
    // also ensure that event2 is different from event1 by comparing the ids
    if (this.id != -1 && this.id != id) {
      // collect the reactions
      reactions(id, merge(this.v1, v2), sync, reacts)
      // reset the sequence
      this.id = -1
    }
  }

  /*
  * Register to the referenced events
  */
  protected override def deploy {
    ev1 += onEvt1
    ev2 += onEvt2
  }

  /*
  * Unregister from the referenced events
  */
  protected override def undeploy {
    ev1 -= onEvt1
    ev2 -= onEvt2
  }

  override def toString = "(" + ev1 + " then " + ev2 + ")"

}

class EventNodeExcept[T,U](accpeted: Event[T], except: Event[U]) extends EventNode[T] {
  
  private val myReacts = new ListBuffer[(() => Unit, Trace)]
  
  private var id = -1
  
  lazy val onAccepted = (id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    myReacts.clear
    // if the id is already set, the except event was received
    if(this.id != id) {
      this.id = id
      reactions(id, v, sync, reacts)
    }
  }
  
  lazy val onExcept = (id: Int, v: U, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) => {
    // the except event is received, set the id to
    if(this.id != id) {
      this.id = id
    } else {
      // remove all my registered reactions
      reacts --= myReacts
      myReacts.clear
    }
  }
  
  override def reactions(id: Int, v: T, sync:Boolean, reacts: ListBuffer[(() => Unit, Trace)]) {
    eventTrace.withValue(this :: eventTrace.value) {
      // collect the reactions of this event
      _reactions.foreach(react => myReacts += (() => react(v), eventTrace.value))
      // collect the reactions of the sinks
      sinks.foreach(sink => sink(id, v, sync, myReacts))
    }
    // add my reactions and my sinks reactions to the global reactions
    reacts ++= myReacts
  }
  
  override def deploy {
    accpeted += onAccepted
    except += onExcept
  }
  
  override def undeploy {
    accpeted -= onAccepted
    except -= onExcept
  }
}

class CausedByFilter(e: Event[_]) extends Function0[Boolean] {
  def apply() = eventTrace.value.contains(e)
  def unary_! = () => !eventTrace.value.contains(e)
}


/*
 * Implementation of an observable list
 */
class VarList[T] private(private val buffer: ListBuffer[T]) extends Iterable[T] {

  def this() = this(new ListBuffer[T])


  // Use list buffer for implementation
  // private val buffer = new ListBuffer[T]

  // delegate the iterator implementation
  override def iterator = buffer.iterator

  /*
   * Add a new element to the list; trigger the corresponding event
   */
  def +=(v: T) = {buffer += v; elementAdded(v)}

  /*
  * Remove an element from the list; trigger the corresponding event
  */
  def -=(v: T) = {buffer -= v; elementRemoved(v)}

  def clear() = {
    buffer.foreach(v => elementRemoved(v))
    buffer.clear()
  }

  def reverse: VarList[T] = new VarList(buffer.reverse)

  //def find(fun: T => Boolean): Option[T] = buffer.find(fun)

  /*
   * A convenience operator creating an event based on the list
   */
  def any[U](evf: T => Event[U]) = new EventNodeExists(this, evf)
  
  /*
  * Events notifying over the changes in the list
  */
  lazy val elementAdded = new ImperativeEvent[T]
  lazy val elementRemoved = new ImperativeEvent[T]
}

object VarList {
  def apply[T](a: Array[T]) = {
    val result = new VarList[T]()
    a.foreach(v => result += v)
    result
  }
  def apply[T](values: Iterable[T]) = {
    val result = new VarList[T]()
    values.foreach(v => result += v)
    result
  }
}

/*
 * Implicit type conversions
 */
object EventsLibConversions {
  // implicit or explicit lifting of methods for instrumentation
  implicit def lift[T, U](body: T => U) = Observable(body)
  // implicitly drop event parameter if not used
  implicit def dropParam[T](ev: Event[T]) = ev.dropParam

  // some implicit conversion for methods which allow to write 
  // instrumented methods in a more intuitive way.
  implicit def toUnitfun[T](f: () => T) = (_: Unit) => f()
  implicit def toTupledFun2[T1,T2,R](f: (T1,T2) => R) = f.tupled
  implicit def toTupledFun3[T1,T2,T3,R](f: (T1,T2,T3) => R) = f.tupled
  implicit def toTupledFun4[T1,T2,T3,T4,R](f: (T1,T2,T3,T4) => R) = f.tupled
  implicit def toTupledFun5[T1,T2,T3,T4,T5,R](f: (T1,T2,T3,T4,T5) => R) = f.tupled
  implicit def toTupledFun6[T1,T2,T3,T4,T5,T6,R](f: (T1,T2,T3,T4,T5,T6) => R) = f.tupled
  implicit def toTupledFun7[T1,T2,T3,T4,T5,T6,T7,R](f: (T1,T2,T3,T4,T5,T6,T7) => R) = f.tupled
  implicit def toTupledFun8[T1,T2,T3,T4,T5,T6,T7,T8,R](f: (T1,T2,T3,T4,T5,T6,T7,T8) => R) = f.tupled
  implicit def toTupledFun9[T1,T2,T3,T4,T5,T6,T7,T8,T9,R](f: (T1,T2,T3,T4,T5,T6,T7,T8,T9) => R) = f.tupled
  implicit def toTupledFun10[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,R](f: (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10) => R) = f.tupled
}

/*
 * Implementation of an observable method
 */
class Observable[T, U](body: T => U) extends (T => U) {
  // before and after, modeled as primitive events
  lazy val before = new ImperativeEvent[T]
  lazy val beforeAsync = new ImperativeAsyncEvent[T]
  lazy val after = new ImperativeEvent[(T,U)]
  lazy val afterAsync = new ImperativeAsyncEvent[(T,U)]

  /*
  * Instrumented method implementation:
  * trigger events before and after the actual method execution
  */
  def apply(t: T): U = {
    before(t)
    beforeAsync(t)
    val res = body(t)
    afterAsync(t,res)
    after(t, res)
    res
  }
}

object Observable {
  def apply[T,U](f: T => U) = new Observable(f)
}
