package scala.events.disjunctjoin

import scala.events.{ Event, EventNode }
import scala.collection.immutable.HashMap
import scala.collection.immutable.HashSet
import scala.collection.mutable.ListBuffer
import scala.collection.immutable.Queue

// the ids for most instances are only useful for debugging
//
// ThreadStore : stores blocked threads
// InEvent : stores incoming events , optianlly their send threads as well
// Chord : a list of events (objects not expressions
//          since they need to be compared between different chords)
// Disjunction : a list of chords  (Methods: markSync and event)
////

object ThreadStore {
  private var _debug = false
  def debug = _debug
  def set_debug(debug_flag: Boolean) = { _debug = debug_flag }
  private var id: Int = 0
  def get_id = synchronized {
    id += 1
    id
  }
}

// the mechanism to block and store threads (for synchronous events)
class ThreadStore(val inEvent: InEvent[_]) {
  val id = ThreadStore.get_id
  val creator = Thread.currentThread
  var task: Option[() => Unit] = None
  var containing: Boolean = false
  var abort: Boolean = false
  def enter {
    if (Thread.currentThread != creator) {
      throw new java.lang.IllegalStateException("Storing another thread than my creator")
    }
    if (ThreadStore.debug) println(this + " entered");
    synchronized {
      containing = true
      while (task.isEmpty && !abort) {
        wait(1000) // 1 sec  probably no timeout needed at all
      }
      if (!abort) task match {
        case None => {}
        case Some(tsk) => {
          tsk()
          if (ThreadStore.debug) println(this + " woke up to execute");
        }
      }
      else {
        if (ThreadStore.debug) println(this + " aborted");
      }
      containing = false
      abort = true // use only once 
    }
  }
  def setTask(tsk: () => Unit) = synchronized {
    if (containing) {
      task = Some(tsk)
      notifyAll
    } else {
      tsk()
      if (ThreadStore.debug) println(this + " executed by same thread");
      abort = true
    }
  }

  def abandon(): Unit = {
    if (!abort) {
      if (ThreadStore.debug) println(this + " abandoned");
      synchronized {
        abort = true
        notifyAll
      }
    }
  }

  override def toString = "ThreadStore[" + id + " for " + inEvent.id + "]{" + (!task.isEmpty) + "} " + creator + " " + Thread.currentThread
}

object InEvent {
  private var _debug = false
  def debug = _debug
  def set_debug(debug_flag: Boolean) = { _debug = debug_flag }
  private var id: Int = 0
  def get_id = synchronized {
    id += 1
    id
  }
  val marked = new ListBuffer[Event[_]]();
  def markNullaryAsync(ev: Event[Unit]) = {
    marked += ev
  }
  def unMarkNullaryAsync(ev: Event[Unit]) = {
    marked -= ev
  }

  def create[T](disjunction: Disjunction, ev: Event[T], idx:Int) = {
    if (marked.contains(ev)) {
      new InNullaryAsyncEvent[T](disjunction, ev,idx)
    } else {
      new InEvent[T](disjunction, ev, idx)
    }
  }
}

// InEvents receive and are able to store event occurrences even with their sending threads
class InEvent[T](val disjunction: Disjunction, val ev: Event[T],val idx:Int) extends EventNode[T] {
  val id = InEvent.get_id
  private var q =  Queue[(T, Option[ThreadStore])]()
  private var enabled: Boolean = false
  private var optimize_async: Boolean = true
  private var cachedIsEmpty: Boolean = true
  private var len:Int=0
  def length = synchronized {
    len
  }
  def isEmpty() = cachedIsEmpty
  protected def setIsEmpty(flag: Boolean) = { cachedIsEmpty = flag }

  def add(id: Int, msg: T, sync: Boolean, reacts: ListBuffer[(() => Unit, Trace)]): Unit = {
    if (InEvent.debug) println(this + "  received id :" +id +" msg:"+ msg + "  sync=" + sync)
    if (sync) {
      // in the optimized case we need to insert different handlers
      _reactions.clear
      _reactions += ((m: T) => enqueueEvent(m,sync))
      reactions(id, msg, sync, reacts)
    } else {
      if (get_optimized_async()) {
        enqueueEvent(msg, sync)
      } else {
        // in the un optimized case this is similar to sync event occurrences
        _reactions.clear
        _reactions += ((m: T) => enqueueEvent(m, sync)) // no threadstore for async events
        reactions(id, msg, sync, reacts)
      }
    }
  }

  def enqueueEvent(msg: T, sync:Boolean): Unit = {
    val tstore=if (sync)  Some(new ThreadStore(this)) else None
    if (InEvent.debug) println(this + "  add to queue :" + msg + "  sync=" + sync)
    synchronized {
      q=q.enqueue(Tuple2(msg, tstore))
      len=len+1
      if (length == 1) {
        setIsEmpty(false)
	disjunction.tryInc(this)
      }
    }
    val handled = disjunction.receive(this)
    if (!handled & sync) {
      tstore match {
	case Some(ts)=>ts.enter
	case None => throw new java.lang.IllegalStateException("Sync and no threadStore!")
      }
    }
  }

  def get(): Option[Tuple2[T, Option[ThreadStore]]] = synchronized {
    if (isEmpty()) {
      None
    } else {
      disjunction.synchronized {
	val (elm,new_q)=q.dequeue
	q=new_q
	len=len-1
        val result = Some(elm)
        if (length == 0) {
          setIsEmpty(true)
        }
        disjunction.tryDec(this)
        result
      }
    }
  }

  def synGet(lst: List[InEvent[_]], collectors: List[() => Option[Tuple2[_, Option[ThreadStore]]]]): List[Tuple2[_, Option[ThreadStore]]] = {
    val getter = (() => get())
    val newCollectors = (getter :: collectors)
    synchronized {
      val nxt = if (lst.isEmpty) null else lst.head
      val tail = if (lst.isEmpty) List() else lst.tail
      if (length == 0) {
	List()
      } else if (lst.isEmpty) {
	// val arg_set=newCollectors.map(x=>x()).reverse
	val arg_set = newCollectors.foldLeft(List[Option[Tuple2[_, Option[ThreadStore]]]]())((z, x) => ((x()) :: z))
	arg_set.map(x => x match {
          case Some(e) => e
          case None => throw new java.lang.IllegalStateException(
	    "not all chords have data? " + collectors.length + " " + arg_set
	  )
	})
      } else if (!nxt.isEmpty()) {
	nxt.synGet(tail, newCollectors)
      } else {
	List()
	}
    }
  }

  def deploy: Unit = { ev += add _ }
  def undeploy: Unit = { ev -= add _ }

  def enable(): Unit = if (!enabled) {
    deploy
    enabled = true
  }

  def get_optimized_async(): Boolean = optimize_async

  def set_optimized_async(flag: Boolean) = {
    optimize_async = flag
  }

  def disable(): Unit = if (enabled) synchronized {
    undeploy
    var ending = false
    while (!ending) { // flush the Q
      get match {
        case None => { ending = true }
        case Some(x) => {
          val ts = x._2
          ts match {
            case None => {}
            case Some(ts) => ts.abandon
          }
        }
      }
    }
    enabled = false
  }

  override def toString = "InEvent[" + disjunction.id + "@" + definer + ";" + id + ",idx:"+idx+"]{" + length + "}"
}

class InNullaryAsyncEvent[T](disjunction: Disjunction, ev: Event[T],idx:Int) extends InEvent[T](disjunction, ev,idx) {
  private var queueLength: Int = 0;
  override def length = queueLength;
  var arg: Option[T] = None

  override def add(id: Int, msg: T, sync: Boolean, reacts: ListBuffer[(() => Unit, Trace)]): Unit = {
    if (sync) {
      // you promised me to be asynchronous only !!!
      throw new IllegalStateException(" " + this + " received a synchronous event!! " + msg)
    } else {
      if (get_optimized_async()) {
        enqueueEvent(msg, sync)
      } else {
        // in the un optimized case this is similar to sync event occurrences
        _reactions.clear
        _reactions += ((m: T) => enqueueEvent(m, false)) // no threadstore for async events
        reactions(id, msg, false, reacts)
      }
    }
  }

  override def enqueueEvent(msg: T, sync:Boolean): Unit = {
  if (InEvent.debug) println(this + "  add to queue :" + msg + "  sync=" + sync)
  val tstore=if (sync) 
	       Some(new ThreadStore(this))
	     else
		 None
    synchronized {
      queueLength += 1;
      if (length == 1) {
        setIsEmpty(false)
        disjunction.tryInc(this)
        arg match {
          case None => arg = Some(msg)
          case Some(_) => {}
        }
      }
    }
    val handled = disjunction.receive(this)
    if (!handled && sync) {
     tstore match {
	case Some(ts)=>throw new IllegalStateException(" " + this + " received a synchronous event!! "+msg)
	case None => throw new java.lang.IllegalStateException("Sync (NullaryAsync!) and no threadStore!")
      }
    }
  }

  override def get(): Option[Tuple2[T, Option[ThreadStore]]] = synchronized {
    if (isEmpty()) {
      None
    } else {
      disjunction.synchronized {
        queueLength -= 1;
        if (length == 0) {
          setIsEmpty(true)
        }
        disjunction.tryDec(this)
        val unitValue: T = arg.get
        Some((unitValue, None))
      }
    }
  }
  override def toString = "InNullaryAsyncEvent[" + disjunction.id+"@"+definer + ";" + id + ",idx:"+idx+"]{" + length + "}"
}

object Chord {
  private var id: Int = 0
  def get_id = synchronized {
    id += 1
    id
  }
  private var _debug = false
  def debug = _debug
  def set_debug(debug_flag: Boolean) = { _debug = debug_flag }

}

// the ImperativeAsyncEvent can be also triggered synchronous
class Chord(val events: List[Event[_]]) extends scala.events.ImperativeAsyncEvent[List[_]] {
  val id = Chord.get_id
  // these two flags are mutually exclusive (the setters should deal with this)
  private var force_sync: Boolean = false
  private var force_async: Boolean = false
  private var enabled:Boolean=false

  var disjunction: Option[Disjunction] = None // the owner
  val (ev2bit, ev2val) = {
    var mp_bit = new HashMap[Event[_], Int]
    var mp_val = new HashMap[Event[_], Tuple2[Int, Int]]
    var pos = events.length
    events.foreach(x => {
      pos -= 1
      mp_bit = mp_bit + ((x, pos))
      val value = (1 << pos).toInt
      mp_val = mp_val + ((x, Tuple2(value, ~value)))
    })
		  (mp_bit, mp_val)
  }

  lazy val eventReceivers = events.map(ev => disjunction match {
    case Some(p) => p.event2receiver(ev)
    case None => throw new java.lang.IllegalStateException("No disjunction specified yet")
  })

  lazy val idx2val={
    val dj=disjunction match {
      case Some(p) => p
      case None => throw new java.lang.IllegalStateException("No disjunction specified yet")
    }
    var idx2val=new Array[Tuple2[Int,Int]](dj.ev2idx.size)
    ev2val.keys.foreach(ev=>{
      idx2val(dj.ev2idx(ev))=(ev2val(ev))
    })
    idx2val
  }

//  lazy val max = (math.pow(2, events.length) - 1).toInt
  val max = ((1 << events.length) - 1).toInt
  var bitmask = 0

  def noMissingEvents = (max == bitmask)

  def tryInc(idx:Int) = {
    bitmask |= idx2val(idx)._1
  }
  
  def tryDec(idx:Int) = {
    bitmask &= idx2val(idx)._2
  }

  def getArgs_and_ThreadStores = {
    val gts = eventReceivers.head.synGet(eventReceivers.tail, List())
  
    // traverse the list just once instead of 1map +  (1map+1filter+1map)
    val (gts1, gts2)=gts.foldRight( (List[Any](),List[ThreadStore]())   )(
				(
				 x:Tuple2[Any,Option[ThreadStore]],
				 z:Tuple2[List[_],List[ThreadStore]]  
				 )=>((
				   (x._1::z._1),
				   x._2 match {
				     case None => z._2
				     case Some (ts) => ts::z._2
				   }))
      )
//    (gts1,gts2)
//   val arg_set = gts.map(x => x._1)
//    val ts_set = gts.map(x => x._2).filter(ts => ts match {
//      case None => false
//      case Some(_) => true
//    }).map(ts => ts.get)
	
//      println("new arg:"+gts1)
//      println("new ts:"+gts2)
//      println("old arg:"+arg_set)
//      println("old ts:"+ts_set)

//(arg_set, ts_set)

 (gts1,gts2)
  }

  // probably synchronized is too secure
  def setDisjunction(d: Disjunction) = synchronized {
    disjunction match {
      case None => disjunction = Some(d)
      case Some(_) => throw new java.lang.IllegalStateException("Changing the disjunction for a Chord is not allowed")
    }
  }

  // execute the handlers of this chord via a ThreadStore
  def exec(args: List[_], tss: List[ThreadStore]): Unit = {
    
    if (tss.isEmpty) {
      val sync=false || force_sync
      if (Chord.debug)
	println(this + " matches (" + args.foldLeft("")((z, x) => z + x.toString + ",") + ") [#sync="+(tss.length)+"] sync:"+sync)
      if (enabled)
	triggerEvent(args, sync) // could be optimized, but shows what is normal
    } else {
      val sync=true && !(force_async)
      val ts = tss.head // maybe we need a better strategy when allowing several sync events
      if (Chord.debug)
	println(this + " matches (" + args.foldLeft("")((z, x) => z + x.toString + ",") + ") [#sync="+(tss.length)+"] sync:"+sync)
      ts.setTask(() => {
	if (enabled)
          triggerEvent(args, sync) // could be optimized, but shows what is normal
        tss.foreach(t => t.abandon()) // only release blocked threads after handling
      })
    }
  }

  // this methods enable experimental behavior
  def set_force_sync(a: Boolean) = {
    force_sync = (!force_async) && a
    (force_sync == a)
  }
  // this two methods enable experimental behavior
  def set_force_async(a: Boolean) = {
    force_async = (!force_sync) && a
    (force_async == a)
  }

  override def undeploy {
    haltAsync()
    super.undeploy
  }
  def disjunction_id = disjunction match {
    case None => "?"
    case Some(x) => x.id
  }
  def enable()=if (!enabled) {
    enabled=true
  }
  def disable()=if (enabled) {
    enabled=false
  }

  override def toString = "Chord[" + disjunction_id + '@' + definer + ";" + id + "](" + events.foldLeft("")((z, x) => z + x.toString + " ") + ")"
  def state = eventReceivers.foldLeft("": String)((z, x) => z + " " + x + ">" + (!x.isEmpty) + "{" + x.length + "}") + "  " + bitmask + "/" + max
}

object Disjunction {
  private var id: Int = 0
  def get_id = synchronized {
    id += 1
    id
  }
  private var _debug = false
  def debug = _debug
  def set_debug(debug_flag: Boolean) = { _debug = debug_flag }
}

class Disjunction(val chords: List[Chord]) {
  protected val definer={val t=Thread.currentThread.getStackTrace.toList
	           val sf=t.lastIndexWhere(x=>x.getClassName.startsWith("scala.events"))+1
		   t(sf).getFileName+":"+t(sf).getLineNumber
	     }
  val id = Disjunction.get_id
  private var deterministic = false

  var enabled = false

  def set_determinism(flag: Boolean) = { deterministic = flag }
  def get_determinism = deterministic

  val event2chords = {
    var ev2chords = new scala.collection.mutable.HashMap[Event[_], List[Chord]]
    chords.foreach(ch => {
      if (ch == null) throw new RuntimeException("null chord? " + chords)
      ch.events.foreach(ev => {
        if (ev == null) throw new RuntimeException("null event? " + ch.events)
        val m = if (ev2chords.contains(ev)) ev2chords(ev) else Nil
        ev2chords(ev) = ch :: m
      })
      ch.setDisjunction(this)
    })
    ev2chords.toMap
  }

  lazy val (evReceivers,event2receiver,idx2receiver,idx2chords,ev2idx) = {
    var receivers: List[InEvent[_]] = Nil
    var event2receiver = new HashMap[Event[_], InEvent[_]]
    val idx2receiver=new Array[InEvent[_]](event2chords.size)
    val idx2chords=new Array[List[Chord]](event2chords.size)
    var ev2idx=new HashMap[Event[_],Int]
    var idx:Int=0
    event2chords.keys.foreach(e => {
      val rec = InEvent.create(this, e, idx)
      receivers = rec :: receivers
      event2receiver = event2receiver + ((e, rec))
      ev2idx = ev2idx + ((e,idx))
      idx2receiver(idx)=rec
      idx2chords(idx)=event2chords(e)
      idx+=1
    })
    (receivers, event2receiver, idx2receiver,idx2chords,ev2idx)
  }

 

  def set_optimized_async(flag: Boolean) = {
    evReceivers.foreach(rec => rec.set_optimized_async(flag))
  }

  // called when receiving an event, decide which chords can be triggered and trigger one
  def receive[T](eventH: InEvent[T]): Boolean = {
    
    if (!enabled)
      false
//    val receivers = event2chords(eventH.ev)
    val receivers = idx2chords(eventH.idx)
    var triggeredChord: Chord = null
    var args: Tuple2[List[_], List[ThreadStore]] = null
    var resolved = false

    synchronized {
      while (!resolved) {
        triggeredChord = null
        args = null
        val candidates = receivers.filter(chrd => chrd.noMissingEvents)
        val crd = chooseChord(candidates)
        crd match {
          case None => {
            triggeredChord = null
            args = null
            resolved = true
          }
          case Some(c) => {
            triggeredChord = c
            args = triggeredChord.getArgs_and_ThreadStores
            if (!args._1.isEmpty) {
              resolved = true
            }
          }
        }
      }
    }
    val arg_set = if (args == null) List() else args._1
    val ts_set = if (args == null) List() else args._2
    
    // only the select chord needs to be locked (not sure even that is needed)
    if (triggeredChord != null && args!=null) {
      triggeredChord.exec(arg_set, ts_set)
      if (Disjunction.debug) println(this + " match, result event was triggered")
      true
    } else {
      if (Disjunction.debug) println(this + " no match")
      false
    }
  }

  // decide wich chord will be triggered to lead to an event
  def chooseChord(candidates: List[Chord]): Option[Chord] = {
    val len = candidates.length
    if (len == 0) {
      None
    } else if (len == 1) { // only one chord is easy
      Some(candidates.head)
    } else if (get_determinism == true) {
      Some(candidates.head)
    } else {
      val choice = util.Random.nextInt(len)
      Some(candidates(choice))
    }
  }

  def tryInc[T](inEv: InEvent[T]) = synchronized {
    if (inEv.length == 1) {
      val idx=inEv.idx
      idx2chords(idx).foreach(crd => crd.tryInc(idx))
//      event2chords(inEv.ev).foreach(crd => crd.tryInc(inEv,idx))
    }
  }

  def tryDec[T](inEv: InEvent[T]) = synchronized {
    if (inEv.isEmpty) {// uses cached info
      val idx=inEv.idx
      idx2chords(idx).foreach(crd => crd.tryDec(idx))
    }
//      event2chords(inEv.ev).foreach(crd => crd.tryDec(inEv,idx))
  }

  // expose the resulting events of each chord
  def event(p: Int) = chords(p)

  def disable(): Unit = if (enabled) {
    if (Disjunction.debug) println(this + " disabled")
    enabled = false
    chords.foreach(ch=>ch.disable())
    evReceivers.foreach(receiver => receiver.disable())
  }

  def enable(): Unit = if (!enabled) {
    if (Disjunction.debug) println(this + " enabled")
    chords.foreach(ch=>ch.enable())
    evReceivers.foreach(receiver => receiver.enable())
    enabled = true
  }
  enable
  override def toString = "Disjunction[" + id + "@" + definer + "](" + chords.foldLeft("")((z, x) => z + x.toString + " ") + ")"
}

