package scala.events.util

import scala.actors.Actor
import scala.collection.mutable.HashMap
import scala.ref.WeakReference
import scala.events.{Event,ImperativeEvent}

object EventTransformer {
  var handler_id:Int=0
  def getNewHandlerId=synchronized {
    handler_id+=1
    handler_id
  }
}

trait EventTransformer {
  gateway:Actor with EventTransformer => 
  private var enabled_events:Int=0
  type IdType=Int
  def haltActor()={} // request to halt the actor 
  def startActor()={} // request to start the actor
  // event exposed to the user
  class ProxyEvent[T](val ev:Option[Event[T]]) extends ImperativeEvent[T] {
    // the registration will be delayed until anything happens
    lazy val id=gateway.register(this,ev)
    override def deploy={
      gateway.enable(id) 
      super.deploy
    }
    override def undeploy={
      super.undeploy
      gateway.disable(id)
    }

    // probably not needed
    def unregister=gateway.unregister(id)

    def trigger_by_gateway(arg:T)=super.apply(arg)

    def asyncTrigger[T]=(arg:T)=>{gateway ! ActorEventMessage(id,arg)}:Unit

    override def apply(arg:T)=if (ev.isEmpty) asyncTrigger() else Unit

    override def toString=getClass.getName+"[#"+id+"]"
  }
 

  // message to send via actor, the id will be used to recoved the Handler
  case class ActorEventMessage[T](val handler_id:Int,arg:T) { }
  case class RequestHalt() {}

  class OriginToActor[T](val id:IdType,val origin_ev:Event[T])  {
    var inputEnabled:Boolean=false
    // the type is lost as message, but the id, will be used to recover it
    lazy val origin_handler=(arg:T)=>{gateway ! ActorEventMessage(id,arg)}:Unit

    def enable=if (!inputEnabled) {
      origin_ev += origin_handler 
      inputEnabled=true
      incrementGatewayUsage()
    }

    def disable=if (inputEnabled) {
      origin_ev -= origin_handler
      inputEnabled=false
      decrementGatewayUsage()
    }
  }


  def incrementGatewayUsage()={
    if (enabled_events==0) {
      startActor()
    }
    enabled_events+=1
  }

  def decrementGatewayUsage()={
    enabled_events-=1
    if (enabled_events==0) 
      gateway.haltActor()
  }


  // receive from origin event, and forwards argument for actor
  // 
  class ActorToEvent[T](val target:WeakReference[ProxyEvent[T]],val origin_ev:Option[Event[T]]=None) {
    val id=EventTransformer.getNewHandlerId
    val origin_to_actor=origin_ev match {
      case Some(origin_ev) => Some(new OriginToActor[T](id,origin_ev))
      case None => None
    }
    def forwardInsideActor(arg:Any) {
      target.get match {
	case Some(re) => re.trigger_by_gateway(arg.asInstanceOf[T])
	case None => gateway.unregister(id)
      }
    }
    def enable:Unit=origin_to_actor match {
      case Some(a2e) => a2e.enable
      case None => {}
    }
    def disable:Unit=origin_to_actor match {
      case Some(a2e) => a2e.disable
      case None => {}
    }

  }

  private val registered_a2e=new HashMap[IdType,ActorToEvent[_]]
  def enable( id:IdType)= registered_a2e(id).enable
  def disable(id:IdType)= registered_a2e(id).disable

  def register[T](ev:ProxyEvent[T],origin_ev:Option[Event[T]])={ 
   val a2e=new ActorToEvent[T](new scala.ref.WeakReference(ev),origin_ev)
    registered_a2e.put(a2e.id,a2e)
    a2e.id
  }
  def unregister(id:IdType)=registered_a2e.get(id) match {
    case Some(h) => {
      h.disable
      registered_a2e.remove(id)
    }
    case None => {}
  }

  // partial function to dispatch in an actor
  val dispatchFunction:PartialFunction[Any,Unit]={
      	case ActorEventMessage(id,arg) => registered_a2e.get(id) match {
	  case Some(handler) => handler.forwardInsideActor(arg)
	  case None => {}  // should not be possible
	}
  }

  def reactEvent(handler: PartialFunction[Any, Unit]):Nothing={
    react(dispatchFunction orElse handler)
  }

  def getProxyEvent[T](ev:Event[T]):ProxyEvent[T]={
    val re=new ProxyEvent(Some(ev))
    re
  }
  def getDelayedEvent[T]={
    val re=new ProxyEvent[T](None)
    re  
  }  

}


class EventGateway extends Actor with EventTransformer {
  private val haltEvent=new ImperativeEvent[Unit]
  private val startEvent=new ImperativeEvent[Unit]
  private var actorRunning:Boolean=false


  def act={
    actorRunning=true
    this.startEvent()
    loopWhile(actorRunning) { 
      reactEvent {
	case RequestHalt => {
	  actorRunning=false
	  this.haltEvent() // trigger event
	}
      }
    }
  }

  override def startActor():Unit={
    // the first event that is deployed, will need the gateway-actor started
    if (!actorRunning)
      this.start
  }


  override def haltActor():Unit= {
    this ! RequestHalt
  }


  def getHaltEvent()=haltEvent

  def getStartEvent()=startEvent
}
